# (c) 2005 Ian Bicking and contributors; written for Paste (http://pythonpaste.org)
# Licensed under the MIT license: http://www.opensource.org/licenses/mit-license.php
"""
Module to find differences over time in a filesystem

Basically this takes a snapshot of a directory, then sees what changes
were made.  The contents of the files are not checked, so you can
detect that the content was changed, but not what the old version of
the file was.
"""

import os
from fnmatch import fnmatch
from datetime import datetime

try:
    # Python 3
    import collections.UserDict as IterableUserDict
except ImportError:
    # Python 2.5-2.7
    from UserDict import IterableUserDict
import operator
import re

__all__ = ['Diff', 'Snapshot', 'File', 'Dir', 'report_expected_diffs',
           'show_diff']

class Diff(object):

    """
    Represents the difference between two snapshots
    """

    def __init__(self, before, after):
        self.before = before
        self.after = after
        self._calculate()

    def _calculate(self):
        before = self.before.data
        after = self.after.data
        self.deleted = {}
        self.updated = {}
        self.created = after.copy()
        for path, f in before.items():
            if path not in after:
                self.deleted[path] = f
                continue
            del self.created[path]
            if f.mtime < after[path].mtime:
                self.updated[path] = after[path]

    def __str__(self):
        return self.report()

    def report(self, header=True, dates=False):
        s = []
        if header:
            s.append('Difference in %s from %s to %s:' %
                     (self.before.base_path,
                      self.before.calculated,
                      self.after.calculated))
        for name, files, show_size in [
            ('created', self.created, True),
            ('deleted', self.deleted, True),
            ('updated', self.updated, True)]:
            if files:
                s.append('-- %s: -------------------' % name)
                files = files.items()
                files.sort()
                last = ''
                for path, f in files:
                    t = '  %s' % _space_prefix(last, path, indent=4,
                                               include_sep=False)
                    last = path
                    if show_size and f.size != 'N/A':
                        t += '  (%s bytes)' % f.size
                    if dates:
                        parts = []
                        if self.before.get(path):
                            parts.append(self.before[path].mtime)
                        if self.after.get(path):
                            parts.append(self.after[path].mtime)
                        t += ' (mtime: %s)' % ('->'.join(map(repr, parts)))
                    s.append(t)
        if len(s) == 1:
            s.append('  (no changes)')
        return '\n'.join(s)

class Snapshot(IterableUserDict):

    """
    Represents a snapshot of a set of files.  Has a dictionary-like
    interface, keyed relative to ``base_path``
    """

    def __init__(self, base_path, files=None, ignore_wildcards=(),
                 ignore_paths=(), ignore_hidden=True):
        self.base_path = base_path
        self.ignore_wildcards = ignore_wildcards
        self.ignore_hidden = ignore_hidden
        self.ignore_paths = ignore_paths
        self.calculated = None
        self.data = files or {}
        if files is None:
            self.find_files()

    ############################################################
    ## File finding
    ############################################################

    def find_files(self):
        """
        Find all the files under the base path, and put them in
        ``self.data``
        """
        self._find_traverse('', self.data)
        self.calculated = datetime.now()

    def _ignore_file(self, fn):
        if fn in self.ignore_paths:
            return True
        if self.ignore_hidden and os.path.basename(fn).startswith('.'):
            return True
        for pat in self.ignore_wildcards:
            if fnmatch(fn, pat):
                return True
        return False

    def _find_traverse(self, path, result):
        full = os.path.join(self.base_path, path)
        if os.path.isdir(full):
            if path:
                # Don't actually include the base path
                result[path] = Dir(self.base_path, path)
            for fn in os.listdir(full):
                fn = os.path.join(path, fn)
                if self._ignore_file(fn):
                    continue
                self._find_traverse(fn, result)
        else:
            result[path] = File(self.base_path, path)

    def __repr__(self):
        return '<%s in %r from %r>' % (
            self.__class__.__name__, self.base_path,
            self.calculated or '(no calculation done)')

    def compare_expected(self, expected, comparison=operator.eq,
                         differ=None, not_found=None,
                         include_success=False):
        """
        Compares a dictionary of ``path: content`` to the
        found files.  Comparison is done by equality, or the
        ``comparison(actual_content, expected_content)`` function given.

        Returns dictionary of differences, keyed by path.  Each
        difference is either noted, or the output of
        ``differ(actual_content, expected_content)`` is given.

        If a file does not exist and ``not_found`` is given, then
        ``not_found(path)`` is put in.
        """
        result = {}
        for path in expected:
            orig_path = path
            path = path.strip('/')
            if path not in self.data:
                if not_found:
                    msg = not_found(path)
                else:
                    msg = 'not found'
                result[path] = msg
                continue
            expected_content = expected[orig_path]
            file = self.data[path]
            actual_content = file.bytes
            if not comparison(actual_content, expected_content):
                if differ:
                    msg = differ(actual_content, expected_content)
                else:
                    if len(actual_content) < len(expected_content):
                        msg = 'differ (%i bytes smaller)' % (
                            len(expected_content) - len(actual_content))
                    elif len(actual_content) > len(expected_content):
                        msg = 'differ (%i bytes larger)' % (
                            len(actual_content) - len(expected_content))
                    else:
                        msg = 'diff (same size)'
                result[path] = msg
            elif include_success:
                result[path] = 'same!'
        return result

    def diff_to_now(self):
        return Diff(self, self.clone())

    def clone(self):
        return self.__class__(base_path=self.base_path,
                              ignore_wildcards=self.ignore_wildcards,
                              ignore_paths=self.ignore_paths,
                              ignore_hidden=self.ignore_hidden)

class File(object):

    """
    Represents a single file found as the result of a command.

    Has attributes:

    ``path``:
        The path of the file, relative to the ``base_path``

    ``full``:
        The full path

    ``stat``:
        The results of ``os.stat``.  Also ``mtime`` and ``size``
        contain the ``.st_mtime`` and ``st_size`` of the stat.

    ``bytes``:
        The contents of the file.

    You may use the ``in`` operator with these objects (tested against
    the contents of the file), and the ``.mustcontain()`` method.
    """

    file = True
    dir = False

    def __init__(self, base_path, path):
        self.base_path = base_path
        self.path = path
        self.full = os.path.join(base_path, path)
        self.stat = os.stat(self.full)
        self.mtime = self.stat.st_mtime
        self.size = self.stat.st_size
        self._bytes = None

    def bytes__get(self):
        if self._bytes is None:
            f = open(self.full, 'rb')
            self._bytes = f.read()
            f.close()
        return self._bytes
    bytes = property(bytes__get)

    def __contains__(self, s):
        return s in self.bytes

    def mustcontain(self, s):
        __tracebackhide__ = True
        bytes = self.bytes
        if s not in bytes:
            print('Could not find %r in:' % s)
            print(bytes)
            assert s in bytes

    def __repr__(self):
        return '<%s %s:%s>' % (
            self.__class__.__name__,
            self.base_path, self.path)

class Dir(File):

    """
    Represents a directory created by a command.
    """

    file = False
    dir = True

    def __init__(self, base_path, path):
        self.base_path = base_path
        self.path = path
        self.full = os.path.join(base_path, path)
        self.size = 'N/A'
        self.mtime = 'N/A'

    def __repr__(self):
        return '<%s %s:%s>' % (
            self.__class__.__name__,
            self.base_path, self.path)

    def bytes__get(self):
        raise NotImplementedError(
            "Directory %r doesn't have content" % self)

    bytes = property(bytes__get)


def _space_prefix(pref, full, sep=None, indent=None, include_sep=True):
    """
    Anything shared by pref and full will be replaced with spaces
    in full, and full returned.

    Example::

        >>> _space_prefix('/foo/bar', '/foo')
        '    /bar'
    """
    if sep is None:
        sep = os.path.sep
    pref = pref.split(sep)
    full = full.split(sep)
    padding = []
    while pref and full and pref[0] == full[0]:
        if indent is None:
            padding.append(' ' * (len(full[0]) + len(sep)))
        else:
            padding.append(' ' * indent)
        full.pop(0)
        pref.pop(0)
    if padding:
        if include_sep:
            return ''.join(padding) + sep + sep.join(full)
        else:
            return ''.join(padding) + sep.join(full)
    else:
        return sep.join(full)

def report_expected_diffs(diffs, colorize=False):
    """
    Takes the output of compare_expected, and returns a string
    description of the differences.
    """
    if not diffs:
        return 'No differences'
    diffs = diffs.items()
    diffs.sort()
    s = []
    last = ''
    for path, desc in diffs:
        t = _space_prefix(last, path, indent=4, include_sep=False)
        if colorize:
            t = color_line(t, 11)
        last = path
        if len(desc.splitlines()) > 1:
            cur_indent = len(re.search(r'^[ ]*', t).group(0))
            desc = indent(cur_indent+2, desc)
            if colorize:
                t += '\n'
                for line in desc.splitlines():
                    if line.strip().startswith('+'):
                        line = color_line(line, 10)
                    elif line.strip().startswith('-'):
                        line = color_line(line, 9)
                    else:
                        line = color_line(line, 14)
                    t += line+'\n'
            else:
                t += '\n' + desc
        else:
            t += ' '+desc
        s.append(t)
    s.append('Files with differences: %s' % len(diffs))
    return '\n'.join(s)

def color_code(foreground=None, background=None):
    """
    0  black
    1  red
    2  green
    3  yellow
    4  blue
    5  magenta (purple)
    6  cyan
    7  white (gray)

    Add 8 to get high-intensity
    """
    if foreground is None and background is None:
        # Reset
        return '\x1b[0m'
    codes = []
    if foreground is None:
        codes.append('[39m')
    elif foreground > 7:
        codes.append('[1m')
        codes.append('[%im' % (22+foreground))
    else:
        codes.append('[%im' % (30+foreground))
    if background is None:
        codes.append('[49m')
    else:
        codes.append('[%im' % (40+background))
    return '\x1b' + '\x1b'.join(codes)

def color_line(line, foreground=None, background=None):
    match = re.search(r'^(\s*)', line)
    return (match.group(1) + color_code(foreground, background)
            + line[match.end():] + color_code())

def indent(indent, text):
    return '\n'.join(
        [' '*indent + l for l in text.splitlines()])

def show_diff(actual_content, expected_content):
    actual_lines = [l.strip() for l in actual_content.splitlines()
                    if l.strip()]
    expected_lines = [l.strip() for l in expected_content.splitlines()
                      if l.strip()]
    if len(actual_lines) == len(expected_lines) == 1:
        return '%r not %r' % (actual_lines[0], expected_lines[0])
    if not actual_lines:
        return 'Empty; should have:\n'+expected_content
    import difflib
    return '\n'.join(difflib.ndiff(actual_lines, expected_lines))



# -*- coding: iso-8859-15 -*-
"""Immutable integer set type.

Integer set class.

Copyright (C) 2006, Heiko Wundram.
Released under the MIT license.
"""
import six

# Version information
# -------------------

__author__ = "Heiko Wundram <me@modelnine.org>"
__version__ = "0.2"
__revision__ = "6"
__date__ = "2006-01-20"


# Utility classes
# ---------------

class _Infinity(object):
    """Internal type used to represent infinity values."""

    __slots__ = ["_neg"]

    def __init__(self,neg):
        self._neg = neg

    def __lt__(self,value):
        if not isinstance(value, _VALID_TYPES):
            return NotImplemented
        return ( self._neg and
                 not ( isinstance(value,_Infinity) and value._neg ) )

    def __le__(self,value):
        if not isinstance(value, _VALID_TYPES):
            return NotImplemented
        return self._neg

    def __gt__(self,value):
        if not isinstance(value, _VALID_TYPES):
            return NotImplemented
        return not ( self._neg or
                     ( isinstance(value,_Infinity) and not value._neg ) )

    def __ge__(self,value):
        if not isinstance(value, _VALID_TYPES):
            return NotImplemented
        return not self._neg

    def __eq__(self,value):
        if not isinstance(value, _VALID_TYPES):
            return NotImplemented
        return isinstance(value,_Infinity) and self._neg == value._neg

    def __ne__(self,value):
        if not isinstance(value, _VALID_TYPES):
            return NotImplemented
        return not isinstance(value,_Infinity) or self._neg != value._neg

    def __repr__(self):
        return "None"

_VALID_TYPES = six.integer_types + (_Infinity,)



# Constants
# ---------

_MININF = _Infinity(True)
_MAXINF = _Infinity(False)


# Integer set class
# -----------------

class IntSet(object):
    """Integer set class with efficient storage in a RLE format of ranges.
    Supports minus and plus infinity in the range."""

    __slots__ = ["_ranges","_min","_max","_hash"]

    def __init__(self,*args,**kwargs):
        """Initialize an integer set. The constructor accepts an unlimited
        number of arguments that may either be tuples in the form of
        (start,stop) where either start or stop may be a number or None to
        represent maximum/minimum in that direction. The range specified by
        (start,stop) is always inclusive (differing from the builtin range
        operator).

        Keyword arguments that can be passed to an integer set are min and
        max, which specify the minimum and maximum number in the set,
        respectively. You can also pass None here to represent minus or plus
        infinity, which is also the default.
        """

        # Special case copy constructor.
        if len(args) == 1 and isinstance(args[0],IntSet):
            if kwargs:
                raise ValueError("No keyword arguments for copy constructor.")
            self._min = args[0]._min
            self._max = args[0]._max
            self._ranges = args[0]._ranges
            self._hash = args[0]._hash
            return

        # Initialize set.
        self._ranges = []

        # Process keyword arguments.
        self._min = kwargs.pop("min",_MININF)
        self._max = kwargs.pop("max",_MAXINF)
        if self._min is None:
            self._min = _MININF
        if self._max is None:
            self._max = _MAXINF

        # Check keyword arguments.
        if kwargs:
            raise ValueError("Invalid keyword argument.")
        if not ( isinstance(self._min, six.integer_types) or self._min is _MININF ):
            raise TypeError("Invalid type of min argument.")
        if not ( isinstance(self._max, six.integer_types) or self._max is _MAXINF ):
            raise TypeError("Invalid type of max argument.")
        if ( self._min is not _MININF and self._max is not _MAXINF and
             self._min > self._max ):
            raise ValueError("Minimum is not smaller than maximum.")
        if isinstance(self._max, six.integer_types):
            self._max += 1

        # Process arguments.
        for arg in args:
            if isinstance(arg, six.integer_types):
                start, stop = arg, arg+1
            elif isinstance(arg,tuple):
                if len(arg) != 2:
                    raise ValueError("Invalid tuple, must be (start,stop).")

                # Process argument.
                start, stop = arg
                if start is None:
                    start = self._min
                if stop is None:
                    stop = self._max

                # Check arguments.
                if not ( isinstance(start, six.integer_types) or start is _MININF ):
                    raise TypeError("Invalid type of tuple start.")
                if not ( isinstance(stop, six.integer_types) or stop is _MAXINF ):
                    raise TypeError("Invalid type of tuple stop.")
                if ( start is not _MININF and stop is not _MAXINF and
                     start > stop ):
                    continue
                if isinstance(stop, six.integer_types):
                    stop += 1
            else:
                raise TypeError("Invalid argument.")

            if start > self._max:
                continue
            elif start < self._min:
                start = self._min
            if stop < self._min:
                continue
            elif stop > self._max:
                stop = self._max
            self._ranges.append((start,stop))

        # Normalize set.
        self._normalize()

    # Utility functions for set operations
    # ------------------------------------

    def _iterranges(self,r1,r2,minval=_MININF,maxval=_MAXINF):
        curval = minval
        curstates = {"r1":False,"r2":False}
        imax, jmax = 2*len(r1), 2*len(r2)
        i, j = 0, 0
        while i < imax or j < jmax:
            if i < imax and ( ( j < jmax and
                                r1[i>>1][i&1] < r2[j>>1][j&1] ) or
                              j == jmax ):
                cur_r, newname, newstate = r1[i>>1][i&1], "r1", not (i&1)
                i += 1
            else:
                cur_r, newname, newstate = r2[j>>1][j&1], "r2", not (j&1)
                j += 1
            if curval < cur_r:
                if cur_r > maxval:
                    break
                yield curstates, (curval,cur_r)
                curval = cur_r
            curstates[newname] = newstate
        if curval < maxval:
            yield curstates, (curval,maxval)

    def _normalize(self):
        self._ranges.sort()
        i = 1
        while i < len(self._ranges):
            if self._ranges[i][0] < self._ranges[i-1][1]:
                self._ranges[i-1] = (self._ranges[i-1][0],
                                     max(self._ranges[i-1][1],
                                         self._ranges[i][1]))
                del self._ranges[i]
            else:
                i += 1
        self._ranges = tuple(self._ranges)
        self._hash = hash(self._ranges)

    def __coerce__(self,other):
        if isinstance(other,IntSet):
            return self, other
        elif isinstance(other, six.integer_types + (tuple,)):
            try:
                return self, self.__class__(other)
            except TypeError:
                # Catch a type error, in that case the structure specified by
                # other is something we can't coerce, return NotImplemented.
                # ValueErrors are not caught, they signal that the data was
                # invalid for the constructor. This is appropriate to signal
                # as a ValueError to the caller.
                return NotImplemented
        elif isinstance(other,list):
            try:
                return self, self.__class__(*other)
            except TypeError:
                # See above.
                return NotImplemented
        return NotImplemented

    # Set function definitions
    # ------------------------

    def _make_function(name,type,doc,pall,pany=None):
        """Makes a function to match two ranges. Accepts two types: either
        'set', which defines a function which returns a set with all ranges
        matching pall (pany is ignored), or 'bool', which returns True if pall
        matches for all ranges and pany matches for any one range. doc is the
        dostring to give this function. pany may be none to ignore the any
        match.

        The predicates get a dict with two keys, 'r1', 'r2', which denote
        whether the current range is present in range1 (self) and/or range2
        (other) or none of the two, respectively."""

        if type == "set":
            def f(self,other):
                coerced = self.__coerce__(other)
                if coerced is NotImplemented:
                    return NotImplemented
                other = coerced[1]
                newset = self.__class__.__new__(self.__class__)
                newset._min = min(self._min,other._min)
                newset._max = max(self._max,other._max)
                newset._ranges = []
                for states, (start,stop) in \
                        self._iterranges(self._ranges,other._ranges,
                                         newset._min,newset._max):
                    if pall(states):
                        if newset._ranges and newset._ranges[-1][1] == start:
                            newset._ranges[-1] = (newset._ranges[-1][0],stop)
                        else:
                            newset._ranges.append((start,stop))
                newset._ranges = tuple(newset._ranges)
                newset._hash = hash(self._ranges)
                return newset
        elif type == "bool":
            def f(self,other):
                coerced = self.__coerce__(other)
                if coerced is NotImplemented:
                    return NotImplemented
                other = coerced[1]
                _min = min(self._min,other._min)
                _max = max(self._max,other._max)
                found = not pany
                for states, (start,stop) in \
                        self._iterranges(self._ranges,other._ranges,_min,_max):
                    if not pall(states):
                        return False
                    found = found or pany(states)
                return found
        else:
            raise ValueError("Invalid type of function to create.")
        try:
            f.func_name = name
        except TypeError:
            pass
        f.func_doc = doc
        return f

    # Intersection.
    __and__ = _make_function("__and__","set",
                             "Intersection of two sets as a new set.",
                             lambda s: s["r1"] and s["r2"])
    __rand__ = _make_function("__rand__","set",
                              "Intersection of two sets as a new set.",
                              lambda s: s["r1"] and s["r2"])
    intersection = _make_function("intersection","set",
                                  "Intersection of two sets as a new set.",
                                  lambda s: s["r1"] and s["r2"])

    # Union.
    __or__ = _make_function("__or__","set",
                            "Union of two sets as a new set.",
                            lambda s: s["r1"] or s["r2"])
    __ror__ = _make_function("__ror__","set",
                             "Union of two sets as a new set.",
                             lambda s: s["r1"] or s["r2"])
    union = _make_function("union","set",
                           "Union of two sets as a new set.",
                           lambda s: s["r1"] or s["r2"])

    # Difference.
    __sub__ = _make_function("__sub__","set",
                             "Difference of two sets as a new set.",
                             lambda s: s["r1"] and not s["r2"])
    __rsub__ = _make_function("__rsub__","set",
                              "Difference of two sets as a new set.",
                              lambda s: s["r2"] and not s["r1"])
    difference = _make_function("difference","set",
                                "Difference of two sets as a new set.",
                                lambda s: s["r1"] and not s["r2"])

    # Symmetric difference.
    __xor__ = _make_function("__xor__","set",
                             "Symmetric difference of two sets as a new set.",
                             lambda s: s["r1"] ^ s["r2"])
    __rxor__ = _make_function("__rxor__","set",
                              "Symmetric difference of two sets as a new set.",
                              lambda s: s["r1"] ^ s["r2"])
    symmetric_difference = _make_function("symmetric_difference","set",
                                          "Symmetric difference of two sets as a new set.",
                                          lambda s: s["r1"] ^ s["r2"])

    # Containership testing.
    __contains__ = _make_function("__contains__","bool",
                                  "Returns true if self is superset of other.",
                                  lambda s: s["r1"] or not s["r2"])
    issubset = _make_function("issubset","bool",
                              "Returns true if self is subset of other.",
                              lambda s: s["r2"] or not s["r1"])
    istruesubset = _make_function("istruesubset","bool",
                                  "Returns true if self is true subset of other.",
                                  lambda s: s["r2"] or not s["r1"],
                                  lambda s: s["r2"] and not s["r1"])
    issuperset = _make_function("issuperset","bool",
                                "Returns true if self is superset of other.",
                                lambda s: s["r1"] or not s["r2"])
    istruesuperset = _make_function("istruesuperset","bool",
                                    "Returns true if self is true superset of other.",
                                    lambda s: s["r1"] or not s["r2"],
                                    lambda s: s["r1"] and not s["r2"])
    overlaps = _make_function("overlaps","bool",
                              "Returns true if self overlaps with other.",
                              lambda s: True,
                              lambda s: s["r1"] and s["r2"])

    # Comparison.
    __eq__ = _make_function("__eq__","bool",
                            "Returns true if self is equal to other.",
                            lambda s: not ( s["r1"] ^ s["r2"] ))
    __ne__ = _make_function("__ne__","bool",
                            "Returns true if self is different to other.",
                            lambda s: True,
                            lambda s: s["r1"] ^ s["r2"])

    # Clean up namespace.
    del _make_function

    # Define other functions.
    def inverse(self):
        """Inverse of set as a new set."""

        newset = self.__class__.__new__(self.__class__)
        newset._min = self._min
        newset._max = self._max
        newset._ranges = []
        laststop = self._min
        for r in self._ranges:
            if laststop < r[0]:
                newset._ranges.append((laststop,r[0]))
                laststop = r[1]
        if laststop < self._max:
            newset._ranges.append((laststop,self._max))
        return newset

    __invert__ = inverse

    # Hashing
    # -------

    def __hash__(self):
        """Returns a hash value representing this integer set. As the set is
        always stored normalized, the hash value is guaranteed to match for
        matching ranges."""

        return self._hash

    # Iterating
    # ---------

    def __len__(self):
        """Get length of this integer set. In case the length is larger than
        2**31 (including infinitely sized integer sets), it raises an
        OverflowError. This is due to len() restricting the size to
        0 <= len < 2**31."""

        if not self._ranges:
            return 0
        if self._ranges[0][0] is _MININF or self._ranges[-1][1] is _MAXINF:
            raise OverflowError("Infinitely sized integer set.")
        rlen = 0
        for r in self._ranges:
            rlen += r[1]-r[0]
        if rlen >= 2**31:
            raise OverflowError("Integer set bigger than 2**31.")
        return rlen

    def len(self):
        """Returns the length of this integer set as an integer. In case the
        length is infinite, returns -1. This function exists because of a
        limitation of the builtin len() function which expects values in
        the range 0 <= len < 2**31. Use this function in case your integer
        set might be larger."""

        if not self._ranges:
            return 0
        if self._ranges[0][0] is _MININF or self._ranges[-1][1] is _MAXINF:
            return -1
        rlen = 0
        for r in self._ranges:
            rlen += r[1]-r[0]
        return rlen

    def __nonzero__(self):
        """Returns true if this integer set contains at least one item."""

        return bool(self._ranges)

    def __iter__(self):
        """Iterate over all values in this integer set. Iteration always starts
        by iterating from lowest to highest over the ranges that are bounded.
        After processing these, all ranges that are unbounded (maximum 2) are
        yielded intermixed."""

        ubranges = []
        for r in self._ranges:
            if r[0] is _MININF:
                if r[1] is _MAXINF:
                    ubranges.extend(([0,1],[-1,-1]))
                else:
                    ubranges.append([r[1]-1,-1])
            elif r[1] is _MAXINF:
                ubranges.append([r[0],1])
            else:
                for val in xrange(r[0],r[1]):
                    yield val
        if ubranges:
            while True:
                for ubrange in ubranges:
                    yield ubrange[0]
                    ubrange[0] += ubrange[1]

    # Printing
    # --------

    def __repr__(self):
        """Return a representation of this integer set. The representation is
        executable to get an equal integer set."""

        rv = []
        for start, stop in self._ranges:
            if ( isinstance(start, six.integer_types) and isinstance(stop, six.integer_types)
                 and stop-start == 1 ):
                rv.append("%r" % start)
            elif isinstance(stop, six.integer_types):
                rv.append("(%r,%r)" % (start,stop-1))
            else:
                rv.append("(%r,%r)" % (start,stop))
        if self._min is not _MININF:
            rv.append("min=%r" % self._min)
        if self._max is not _MAXINF:
            rv.append("max=%r" % self._max)
        return "%s(%s)" % (self.__class__.__name__,",".join(rv))

if __name__ == "__main__":
    # Little test script demonstrating functionality.
    x = IntSet((10,20),30)
    y = IntSet((10,20))
    z = IntSet((10,20),30,(15,19),min=0,max=40)
    print(x)
    print(x&110)
    print(x|110)
    print(x^(15,25))
    print(x-12)
    print(12 in x)
    print(x.issubset(x))
    print(y.issubset(x))
    print(x.istruesubset(x))
    print(y.istruesubset(x))
    for val in x:
        print(val)
    print(x.inverse())
    print(x == z)
    print(x == y)
    print(x != y)
    print(hash(x))
    print(hash(z))
    print(len(x))
    print(x.len())



# (c) 2005 Clark C. Evans and contributors
# This module is part of the Python Paste Project and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php
# Some of this code was funded by: http://prometheusresearch.com
"""
Date, Time, and Timespan Parsing Utilities

This module contains parsing support to create "human friendly"
``datetime`` object parsing.  The explicit goal of these routines is
to provide a multi-format date/time support not unlike that found in
Microsoft Excel.  In most approaches, the input is very "strict" to
prevent errors -- however, this approach is much more liberal since we
are assuming the user-interface is parroting back the normalized value
and thus the user has immediate feedback if the data is not typed in
correctly.

  ``parse_date`` and ``normalize_date``

     These functions take a value like '9 jan 2007' and returns either an
     ``date`` object, or an ISO 8601 formatted date value such
     as '2007-01-09'.  There is an option to provide an Oracle database
     style output as well, ``09 JAN 2007``, but this is not the default.

     This module always treats '/' delimiters as using US date order
     (since the author's clients are US based), hence '1/9/2007' is
     January 9th.  Since this module treats the '-' as following
     European order this supports both modes of data-entry; together
     with immediate parroting back the result to the screen, the author
     has found this approach to work well in pratice.

  ``parse_time`` and ``normalize_time``

     These functions take a value like '1 pm' and returns either an
     ``time`` object, or an ISO 8601 formatted 24h clock time
     such as '13:00'.  There is an option to provide for US style time
     values, '1:00 PM', however this is not the default.

  ``parse_datetime`` and ``normalize_datetime``

     These functions take a value like '9 jan 2007 at 1 pm' and returns
     either an ``datetime`` object, or an ISO 8601 formatted
     return (without the T) such as '2007-01-09 13:00'. There is an
     option to provide for Oracle / US style, '09 JAN 2007 @ 1:00 PM',
     however this is not the default.

  ``parse_delta`` and ``normalize_delta``

     These functions take a value like '1h 15m' and returns either an
     ``timedelta`` object, or an 2-decimal fixed-point
     numerical value in hours, such as '1.25'.  The rationale is to
     support meeting or time-billing lengths, not to be an accurate
     representation in mili-seconds.  As such not all valid
     ``timedelta`` values will have a normalized representation.

"""
from datetime import timedelta, time, date
from time import localtime

__all__ = ['parse_timedelta', 'normalize_timedelta',
           'parse_time', 'normalize_time',
           'parse_date', 'normalize_date']

def _number(val):
    try:
        return int(val)
    except:
        return None

#
# timedelta
#
def parse_timedelta(val):
    """
    returns a ``timedelta`` object, or None
    """
    if not val:
        return None
    val = val.lower()
    if "." in val:
        val = float(val)
        return timedelta(hours=int(val), minutes=60*(val % 1.0))
    fHour = ("h" in val or ":" in val)
    fMin  = ("m" in val or ":" in val)
    for noise in "minu:teshour()":
        val = val.replace(noise, ' ')
    val = val.strip()
    val = val.split()
    hr = 0.0
    mi = 0
    val.reverse()
    if fHour:
        hr = int(val.pop())
    if fMin:
        mi = int(val.pop())
    if len(val) > 0 and not hr:
        hr = int(val.pop())
    return timedelta(hours=hr, minutes=mi)

def normalize_timedelta(val):
    """
    produces a normalized string value of the timedelta

    This module returns a normalized time span value consisting of the
    number of hours in fractional form. For example '1h 15min' is
    formatted as 01.25.
    """
    if type(val) == str:
        val = parse_timedelta(val)
    if not val:
        return ''
    hr = val.seconds/3600
    mn = (val.seconds % 3600)/60
    return "%d.%02d" % (hr, mn * 100/60)

#
# time
#
def parse_time(val):
    if not val:
        return None
    hr = mi = 0
    val = val.lower()
    amflag = (-1 != val.find('a'))  # set if AM is found
    pmflag = (-1 != val.find('p'))  # set if PM is found
    for noise in ":amp.":
        val = val.replace(noise, ' ')
    val = val.split()
    if len(val) > 1:
        hr = int(val[0])
        mi = int(val[1])
    else:
        val = val[0]
        if len(val) < 1:
            pass
        elif 'now' == val:
            tm = localtime()
            hr = tm[3]
            mi = tm[4]
        elif 'noon' == val:
            hr = 12
        elif len(val) < 3:
            hr = int(val)
            if not amflag and not pmflag and hr < 7:
                hr += 12
        elif len(val) < 5:
            hr = int(val[:-2])
            mi = int(val[-2:])
        else:
            hr = int(val[:1])
    if amflag and hr >= 12:
        hr = hr - 12
    if pmflag and hr < 12:
        hr = hr + 12
    return time(hr, mi)

def normalize_time(value, ampm):
    if not value:
        return ''
    if type(value) == str:
        value = parse_time(value)
    if not ampm:
        return "%02d:%02d" % (value.hour, value.minute)
    hr = value.hour
    am = "AM"
    if hr < 1 or hr > 23:
        hr = 12
    elif hr >= 12:
        am = "PM"
        if hr > 12:
            hr = hr - 12
    return "%02d:%02d %s" % (hr, value.minute, am)

#
# Date Processing
#

_one_day = timedelta(days=1)

_str2num = {'jan':1, 'feb':2, 'mar':3, 'apr':4,  'may':5, 'jun':6,
            'jul':7, 'aug':8, 'sep':9, 'oct':10, 'nov':11, 'dec':12 }

def _month(val):
    for (key, mon) in _str2num.items():
        if key in val:
            return mon
    raise TypeError("unknown month '%s'" % val)

_days_in_month = {1: 31, 2: 28, 3: 31, 4: 30, 5: 31, 6: 30,
                  7: 31, 8: 31, 9: 30, 10: 31, 11: 30, 12: 31,
                  }
_num2str = {1: 'Jan', 2: 'Feb', 3: 'Mar', 4: 'Apr', 5: 'May', 6: 'Jun',
            7: 'Jul', 8: 'Aug', 9: 'Sep', 10: 'Oct', 11: 'Nov', 12: 'Dec',
            }
_wkdy = ("mon", "tue", "wed", "thu", "fri", "sat", "sun")

def parse_date(val):
    if not(val):
        return None
    val = val.lower()
    now = None

    # optimized check for YYYY-MM-DD
    strict = val.split("-")
    if len(strict) == 3:
        (y, m, d) = strict
        if "+" in d:
            d = d.split("+")[0]
        if " " in d:
            d = d.split(" ")[0]
        try:
            now = date(int(y), int(m), int(d))
            val = "xxx" + val[10:]
        except ValueError:
            pass

    # allow for 'now', 'mon', 'tue', etc.
    if not now:
        chk = val[:3]
        if chk in ('now','tod'):
            now = date.today()
        elif chk in _wkdy:
            now = date.today()
            idx = list(_wkdy).index(chk) + 1
            while now.isoweekday() != idx:
                now += _one_day

    # allow dates to be modified via + or - /w number of days, so
    # that now+3 is three days from now
    if now:
        tail = val[3:].strip()
        tail = tail.replace("+"," +").replace("-"," -")
        for item in tail.split():
            try:
                days = int(item)
            except ValueError:
                pass
            else:
                now += timedelta(days=days)
        return now

    # ok, standard parsing
    yr = mo = dy = None
    for noise in ('/', '-', ',', '*'):
        val = val.replace(noise, ' ')
    for noise in _wkdy:
        val = val.replace(noise, ' ')
    out = []
    last = False
    ldig = False
    for ch in val:
        if ch.isdigit():
            if last and not ldig:
               out.append(' ')
            last = ldig = True
        else:
            if ldig:
                out.append(' ')
                ldig = False
            last = True
        out.append(ch)
    val = "".join(out).split()
    if 3 == len(val):
        a = _number(val[0])
        b = _number(val[1])
        c = _number(val[2])
        if len(val[0]) == 4:
            yr = a
            if b:  # 1999 6 23
                mo = b
                dy = c
            else:  # 1999 Jun 23
                mo = _month(val[1])
                dy = c
        elif a is not None and a > 0:
            yr = c
            if len(val[2]) < 4:
                raise TypeError("four digit year required")
            if b: # 6 23 1999
                dy = b
                mo = a
            else: # 23 Jun 1999
                dy = a
                mo = _month(val[1])
        else: # Jun 23, 2000
            dy = b
            yr = c
            if len(val[2]) < 4:
                raise TypeError("four digit year required")
            mo = _month(val[0])
    elif 2 == len(val):
        a = _number(val[0])
        b = _number(val[1])
        if a is not None and a > 999:
            yr = a
            dy = 1
            if b is not None and b > 0: # 1999 6
                mo = b
            else: # 1999 Jun
                mo = _month(val[1])
        elif a is not None and a > 0:
            if b is not None and b > 999: # 6 1999
                mo = a
                yr = b
                dy = 1
            elif b is not None and b > 0: # 6 23
                mo = a
                dy = b
            else: # 23 Jun
                dy = a
                mo = _month(val[1])
        else:
            if b > 999: # Jun 2001
                yr = b
                dy = 1
            else:  # Jun 23
                dy = b
            mo = _month(val[0])
    elif 1 == len(val):
        val = val[0]
        if not val.isdigit():
            mo = _month(val)
            if mo is not None:
                dy = 1
        else:
            v = _number(val)
            val = str(v)
            if 8 == len(val): # 20010623
                yr = _number(val[:4])
                mo = _number(val[4:6])
                dy = _number(val[6:])
            elif len(val) in (3,4):
                if v is not None and v > 1300: # 2004
                    yr = v
                    mo = 1
                    dy = 1
                else:        # 1202
                    mo = _number(val[:-2])
                    dy = _number(val[-2:])
            elif v < 32:
                dy = v
            else:
                raise TypeError("four digit year required")
    tm = localtime()
    if mo is None:
        mo = tm[1]
    if dy is None:
        dy = tm[2]
    if yr is None:
        yr = tm[0]
    return date(yr, mo, dy)

def normalize_date(val, iso8601=True):
    if not val:
        return ''
    if type(val) == str:
        val = parse_date(val)
    if iso8601:
        return "%4d-%02d-%02d" % (val.year, val.month, val.day)
    return "%02d %s %4d" % (val.day, _num2str[val.month], val.year)
